/*****************************************************************************\
 *
 * chunkedLOD.cpp
 *
 * chunkedLOD - implementation of chunkedLOD class
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/ 

#include <Inventor/SbLinear.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoLOD.h>
#include "chunkedLOD.h"

#include <Inventor/nodes/SoPointSet.h>


//! callback funtion for load all data from model
void chunkedLOD::triangle_cb(void * userdata, SoCallbackAction * action, const SoPrimitiveVertex * v1, const SoPrimitiveVertex * v2, const SoPrimitiveVertex * v3)
{
	//take pointer to my owner
	chunkedLOD * owner = (chunkedLOD*) userdata;

	const SbVec3f vtx[] = { v1->getPoint(), v2->getPoint(), v3->getPoint() };
	const SbVec4f tex[] = { v1->getTextureCoords(), v2->getTextureCoords(), v3->getTextureCoords() };

	// transf. matice
	const SbMatrix mm = action->getModelMatrix();

	// tr. - vrcholy v glob. sour.
	SbVec3f vx[3];
	int tri_index[3];
	int i;
	float x,y,z,w;

	for (i=0; i < 3; i++) 
	{ 
		mm.multVecMatrix(vtx[i], vx[i]);
	}


	for (i=0; i < 3; i++)
	{		
		tri_index[i] = owner->coord3_tmp->findPoint(vx[i]);
		if (tri_index[i] == -1) tri_index[i] = owner->coord3_tmp->addPoint(vx[i],NULL);
		if (owner->tex_coord2->point.getNum() - 1 < tri_index[i])
		{
			owner->tex_coord2->point.setNum(tri_index[i] + 1);
		}
		tex[i].getValue(x,y,z,w);
		owner->tex_coord2->point.set1Value(tri_index[i],x,y);
	}

	int indices[] = { tri_index[0], tri_index[1], tri_index[2], -1 };

	int oldsize = owner->ifs->coordIndex.getNum();
	owner->ifs->coordIndex.setNum(oldsize + 4);
	owner->ifs->coordIndex.setValues(oldsize, 4, indices);
}

//! callback function for load texture from input file
SoCallbackAction::Response chunkedLOD::texture_cb(void * userdata, SoCallbackAction * action, const SoNode * tex)
{
	//take pointer to my owner
	chunkedLOD * owner = (chunkedLOD*) userdata;
	SoTexture2 * tex2;

	
	if (tex->getTypeId() == SoTexture2::getClassTypeId())
	{
		tex2 = (SoTexture2*) tex;
		owner->texture2->filename = tex2->filename;
		owner->texture2->wrapS = tex2->wrapS;
		owner->texture2->wrapT = tex2->wrapT;
		owner->texture2->model = tex2->model;
		owner->texture2->blendColor = tex2->blendColor;
	}

	return SoCallbackAction::CONTINUE;
}

//! quicksort for sorting triangles to elemination
void chunkedLOD::QuickSort(SoMFFloat * tmp_distance, SoMFInt32 * tmp_delete, SoMFInt32 * tmp_join, int left, int right)
{
	int i,j;
	float pm;
	float tmp_f;
	int tmp_i;

	i = left;
	j = right;
	pm = (*tmp_distance)[(i + j) / 2];
	do
	{
		while ((*tmp_distance)[i] < pm)
		{
			i++;
		}
		while ((*tmp_distance)[j] > pm)
		{
			j--;
		}
		if (i <= j)
		{
			tmp_f = (*tmp_distance)[i];
			tmp_distance->set1Value(i,(*tmp_distance)[j]);
			tmp_distance->set1Value(j,tmp_f);

			tmp_i = (*tmp_delete)[i];
			tmp_delete->set1Value(i,(*tmp_delete)[j]);
			tmp_delete->set1Value(j,tmp_i);

			tmp_i = (*tmp_join)[i];
			tmp_join->set1Value(i,(*tmp_join)[j]);
			tmp_join->set1Value(j,tmp_i);

			i++;
			j--;
		}
	}
	while (i < j);

	if (left < j) QuickSort(tmp_distance, tmp_delete, tmp_join, left,j);
	if (right > i) QuickSort(tmp_distance, tmp_delete, tmp_join,i,right);
}

//! constructor
chunkedLOD::chunkedLOD()
{
	//init
	root = NULL;
	originalModel = new SoSeparator;
	coord3 = new SoCoordinate3;
	coord3->point.setNum(0);
	ifs = new SoIndexedFaceSet;
	ifs->coordIndex.setNum(0);
	tex_coord2 = new SoTextureCoordinate2;
	tex_coord2->point.setNum(0);
	coord3_tmp = new SbBSPTree;
	texture2 = new SoTexture2();
}

//! destructor
chunkedLOD::~chunkedLOD()
{

}

//! load model from iv file
bool chunkedLOD::LoadModel(const char * filename)
{
	SoInput in;
  SoSeparator * objectRoot = new SoSeparator;

	printf("Loading %s ...",filename);

	//load file
	if (in.openFile(filename))
	{
		objectRoot = SoDB::readAll(&in);
		objectRoot->ref();

		//callback funtion
		SoCallbackAction ca;
		ca.addTriangleCallback(SoShape::getClassTypeId(), triangle_cb, this);
		ca.addPostCallback(SoTexture2::getClassTypeId(), texture_cb, this);
		ca.apply(objectRoot);

		//copy triangles from BSPtree to coordinate3
		coord3->point.setNum(coord3_tmp->numPoints());
		for (int i = 0; i < coord3_tmp->numPoints(); i++)
		{
			coord3->point.set1Value(i,coord3_tmp->getPoint(i));
		}

		objectRoot->unref();
		AddTexture(originalModel);
		originalModel->addChild(coord3);
		originalModel->addChild(tex_coord2);
		originalModel->addChild(ifs);
		originalModel->ref();

		printf(" Done\n");
		return true;
	}
	else
	{
		printf(" Failed\n");
		return false;
	}
}

//! save model to iv file
void chunkedLOD::SaveModel(SoSeparator * model, const char * filename)
{
  SoOutput output;
	printf("Saving %s ...",filename);
  output.openFile(filename);
  SoWriteAction * writer = new SoWriteAction(&output);
  writer->apply(model);
  delete writer;
	printf("Done\n");
}

//! find minimal and maximal values from model
void chunkedLOD::FindMinMax(SoSeparator * origin, float &minx, float &miny, float &maxx, float &maxy)
{
	//pointers to origin model
	SoCoordinate3 * tmp_coord3 = NULL;
	float x, y, z;

	//search for coordinates
  SoSearchAction sa;
  sa.setType(SoCoordinate3::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	SoFullPath * p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
	}
	else
	{
		tmp_coord3 = (SoCoordinate3*) p->getTail();
		minx = 999999.0f;
		miny = 999999.0f;
		maxx = -999999.0f;
		maxy = -999999.0f;
		for(int i = 0; i < tmp_coord3->point.getNum(); i++)
		{
			tmp_coord3->point[i].getValue(x,y,z);
			if (x < minx) minx = x;
			if (x > maxx) maxx = x;
			if (z < miny) miny = z;
			if (z > maxy) maxy = z;

		}
	}
}

//! add texture to model
void chunkedLOD::AddTexture(SoSeparator * origin)
{
	origin->addChild(texture2);
}

//! find and eliminate bad triangles
void chunkedLOD::CorrectTriangles(SoCoordinate3 * tmp_coord, SoIndexedFaceSet * tmp_ifs)
{
	int i;
	SbLine l1,l2;
	SbVec3f vec1, vec2;
	for(i = 0; i < tmp_ifs->coordIndex.getNum() / 4; i++)
	{
		/*
		l1.setValue(tmp_coord->point[tmp_ifs->coordIndex[i * 4]],tmp_coord->point[tmp_ifs->coordIndex[i * 4 + 1]]);
		l2.setValue(tmp_coord->point[tmp_ifs->coordIndex[i * 4]],tmp_coord->point[tmp_ifs->coordIndex[i * 4 + 2]]);
		vec1 = l1.getDirection();
		vec2 = l2.getDirection();
		if (vec1.equals(vec2,0.000000001f) == 1)
		{
			tmp_ifs->coordIndex.deleteValues(i*4,4);
			i--;
		}
		else
		{
			vec2.negate();
			if (vec1.equals(vec2,0.000000001f) == 1)
			{
				tmp_ifs->coordIndex.deleteValues(i*4,4);
				i--;
			}
			
		}
		*/
		l1.setValue(tmp_coord->point[tmp_ifs->coordIndex[i * 4]],tmp_coord->point[tmp_ifs->coordIndex[i * 4 + 1]]);
		vec1 = tmp_coord->point[tmp_ifs->coordIndex[i * 4 + 2]];
		vec2 = l1.getClosestPoint(vec1);
		if (vec2.equals(tmp_coord->point[tmp_ifs->coordIndex[i * 4 + 2]],0.000000001f))
		{
				tmp_ifs->coordIndex.deleteValues(i*4,4);
				i--;
		}

	}
}

//! cut segment (NON rectangular cutting)
SoSeparator* chunkedLOD::CutSegment(SoSeparator *origin, float x1, float y1, float x2, float y2, SbBSPTree * border)
{
	SoSeparator * dest = new SoSeparator;

	SoCoordinate3 * dest_coord3 = new SoCoordinate3;
	SbBSPTree * dest_bsp_coord3 = new SbBSPTree(16,0);
	SoIndexedFaceSet * dest_ifs = new SoIndexedFaceSet;
	SoTextureCoordinate2 * dest_tex_coord2 = new SoTextureCoordinate2;

	//pointers to origin model
	SoCoordinate3 * tmp_coord3 = NULL;
	SoIndexedFaceSet * tmp_ifs = NULL;
	SoTextureCoordinate2 * tmp_tex = NULL;

	//clearing dest structures
	dest->ref();
	dest_coord3->ref();
	dest_coord3->point.setNum(0);
	dest_ifs->ref();
	dest_ifs->coordIndex.setNum(0);


	//search for coordinates
  SoSearchAction sa;
  sa.setType(SoCoordinate3::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	SoFullPath * p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		delete dest_bsp_coord3;
		return NULL;
	}
	else
	{
		tmp_coord3 = (SoCoordinate3*) p->getTail();
	}

	//search for face set
  sa.setType(SoIndexedFaceSet::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		delete dest_bsp_coord3;
		return NULL;
	}
	else
	{
		tmp_ifs = (SoIndexedFaceSet*) p->getTail();
	}

	//search for texture coordinate
  sa.setType(SoTextureCoordinate2::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		delete dest_bsp_coord3;
		return NULL;
	}
	else
	{
		tmp_tex = (SoTextureCoordinate2*) p->getTail();
	}


	float x,y,z;
	float center_x,center_y;
	SbVec3f tmp_vec;
	int i,j;
	int a,b,c;
	int tri_center_in;

	SbVec3f tri[3];
	int tri_in[3];
	for (i = 0; i < tmp_ifs->coordIndex.getNum() / 4; i++)
	//for all trinangles
	{

		for (j = 0; j < 3; j++)
		{
			tri[j] = tmp_coord3->point[tmp_ifs->coordIndex[i*4 + j]];
			tri[j].getValue(x,y,z);
			tri_in[j] = 0;
			if ((x > x1) && (x < x2) && (z > y1) && (z < y2)) //point is in new segment
			{
				//point is in segment
				tri_in[j] = 1;
			}
		}
		center_x = 0.0;
		center_y = 0.0;
		for (c = 0; c < 3; c++)
		{
			//calculate center of triangle
			tri[c].getValue(x,y,z);
			center_x = center_x + x;
			center_y = center_y + z;
		}
		center_x = center_x / 3.0f;
		center_y = center_y / 3.0f;

		tri_center_in = 0;
		if ((center_x >= x1) && (center_x <= x2) && (center_y >= y1) && (center_y <= y2)) //point is in new segment
		{
			//center of triangle is in segment => triangle is in segment
			tri_center_in = 1;

			//add triangle to segment
			for (c = 0; c < 3; c++)
			{
				//make place for new coordIndex
				dest_ifs->coordIndex.setNum(dest_ifs->coordIndex.getNum() + 1);
				//fill with value from dest_coord3 (make new if not found)
				a = dest_bsp_coord3->findPoint(tri[c]);
				if (a == -1) a = dest_bsp_coord3->addPoint(tri[c],NULL);
				dest_ifs->coordIndex.set1Value(dest_ifs->coordIndex.getNum()-1, a);

				//add texture coordinates
				if (dest_tex_coord2->point.getNum() - 1 < a)
				{
					dest_tex_coord2->point.setNum(a + 1);
				}
				tmp_tex->point[tmp_ifs->coordIndex[i*4 + c]].getValue(x,y);
				dest_tex_coord2->point.set1Value(a,x,y);
			}
			//finish the triangle
			dest_ifs->coordIndex.setNum(dest_ifs->coordIndex.getNum() + 1);
			dest_ifs->coordIndex.set1Value(dest_ifs->coordIndex.getNum()-1, -1);
		}

		//test points from triangle for borders
		switch ((tri_in[0] + tri_in[1] + tri_in[2]))
		{
		case 0:
			if (tri_center_in == 1)
			{
				for (c = 0; c < 3; c++)
				{
					//add point to border
					b = border->findPoint(tri[c]);
					if (b == -1) border->addPoint(tri[c], NULL);
				}

			}
			break;
		case 1:
		case 2:
			//one or two points are in => two or one point is out
			if (tri_center_in == 0)
			{
				for (c = 0; c < 3; c++)
				{
					if (tri_in[c] == 1)
					{
						//add point to border
						b = border->findPoint(tri[c]);
						if (b == -1) border->addPoint(tri[c], NULL);
					}
				}

			}
			else
			{
				for (c = 0; c < 3; c++)
				{
					if (tri_in[c] == 0)
					{
						//add point to border
						b = border->findPoint(tri[c]);
						if (b == -1) border->addPoint(tri[c], NULL);
					}
				}
			}
			break;
		}
	} //for all triangles

	//add structures to final node
	dest_coord3->point.setNum(dest_bsp_coord3->numPoints());
	for (i = 0; i < dest_bsp_coord3->numPoints(); i++)
	{
		dest_coord3->point.set1Value(i,dest_bsp_coord3->getPoint(i));
	}
	delete dest_bsp_coord3;

	dest->addChild(dest_coord3);
	dest->addChild(dest_tex_coord2);
	dest->addChild(dest_ifs);

	return dest;
}

//! reduce number of triangles
void chunkedLOD::SimplifySegment(SoSeparator *origin, SbBSPTree * border, int reduce)
{
	SoCoordinate3 * tmp_coord3;
	SoIndexedFaceSet * tmp_ifs;
	SoTextureCoordinate2 * tmp_tex = NULL;

	SoMFFloat * tmp_distance = NULL;
	SoMFInt32 * tmp_delete = NULL;
	SoMFInt32 * tmp_join = NULL;
	SoMFInt32 * tmp_coord_delete = NULL;
	SoMFInt32 * tmp_coord_join = NULL;
	SoMFInt32 * tmp_coord_joined = NULL;

	int i,j;
	int start_triangles;
	int coord_delete = 0;
	int coord_join = 0;
	float min_distance;
	float distance;
	float x,y,z;
	float xx,yy,zz;
	int tri[3];
	int cycle;
	SbVec3f tria[3];

	SbVec3f tmp_point;

	//search for coordinates
  SoSearchAction sa;
  sa.setType(SoCoordinate3::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	SoFullPath * p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		return;
	}
	else
	{
		tmp_coord3 = (SoCoordinate3*) p->getTail();
	}

	//search for face set
  sa.setType(SoIndexedFaceSet::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		return;
	}
	else
	{
		tmp_ifs = (SoIndexedFaceSet*) p->getTail();
	}

	//search for texture coordinate
  sa.setType(SoTextureCoordinate2::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
  sa.setSearchingAll(TRUE);
  sa.apply(origin);

	p = (SoFullPath*) sa.getPath();
	if (p == NULL) 
	{
		return;
	}
	else
	{
		tmp_tex = (SoTextureCoordinate2*) p->getTail();
	}

	start_triangles = tmp_ifs->coordIndex.getNum() / 4;

	tmp_distance = new SoMFFloat;
	tmp_distance->setNum(0);
	tmp_delete = new SoMFInt32;
	tmp_delete->setNum(0);
	tmp_join = new SoMFInt32;
	tmp_join->setNum(0);
	tmp_coord_delete = new SoMFInt32;
	tmp_coord_delete->setNum(tmp_coord3->point.getNum());
	tmp_coord_join = new SoMFInt32;
	tmp_coord_join->setNum(tmp_coord3->point.getNum());
	tmp_coord_joined = new SoMFInt32;
	tmp_coord_joined->setNum(tmp_coord3->point.getNum());
	
	int coord_found;
	int num_for_delete = 0;
	for (i = 0; i < tmp_ifs->coordIndex.getNum() / 4; i++)
	{
		for (j = 0; j < 3; j++)
		{
			tria[j] = tmp_coord3->point[tmp_ifs->coordIndex[i * 4 + j]];
		}
		min_distance = 999999.0f;
		coord_found = 0;
		for (j = 0; j < 3; j++)
		{
			if (border->findPoint(tria[j]) == -1)
			{
				tria[j].getValue(x,y,z);
				tria[(j + 1) % 3].getValue(xx,yy,zz);
				distance = (x - xx) * (x - xx) + (y - yy) * (y - yy);
				distance = distance + (z - zz) * (z - zz);
				if (distance < min_distance)
				{
					min_distance = distance;
					coord_delete = tmp_ifs->coordIndex[i * 4 + j];
					coord_join = tmp_ifs->coordIndex[i * 4 + ((j + 1) % 3)];
					coord_found = 1;
				}
			}
		}
		if (coord_found == 1)
		{
			num_for_delete++;
			tmp_distance->setNum(num_for_delete);
			tmp_distance->set1Value(num_for_delete-1,min_distance);
			tmp_delete->setNum(num_for_delete);
			tmp_delete->set1Value(num_for_delete-1,coord_delete);
			tmp_join->setNum(num_for_delete);
			tmp_join->set1Value(num_for_delete-1,coord_join);
		}
	}
	if (tmp_delete->getNum() != 0)
		QuickSort(tmp_distance,tmp_delete, tmp_join, 0,tmp_distance->getNum() - 1);

	//init
	for (i = 0; i < tmp_coord3->point.getNum(); i++)
	{
		tmp_coord_delete->set1Value(i,-1);
		tmp_coord_join->set1Value(i,-1);
		tmp_coord_joined->set1Value(i,-1);
	}

	//coord elimination => fill tmp_coord_delete and tmp_coord_join
	int deleted_triangles = 0;
	cycle = 0;
	while ((cycle < tmp_delete->getNum()) && (deleted_triangles < start_triangles * (reduce) / 100))
	{
		coord_delete = (*tmp_delete)[cycle];
		coord_join = (*tmp_join)[cycle];

		if (((*tmp_coord_delete)[coord_delete] == -1) && ((*tmp_coord_delete)[coord_join] == -1) && ((*tmp_coord_joined)[coord_delete] == -1) && ((*tmp_coord_joined)[coord_join] == -1))
		{
			tmp_coord_delete->set1Value(coord_delete, 1);
			tmp_coord_join->set1Value(coord_delete, coord_join);
			tmp_coord_joined->set1Value(coord_join, 1);
			deleted_triangles+=2;
		}
		cycle++;
	}

	//calculate coord diferencies
	int diff = 0;
	for (i = 0; i < tmp_coord3->point.getNum(); i++)
	{
		if ((*tmp_coord_delete)[i] == 1) 
		{
			tmp_coord_delete->set1Value(i,diff);
			diff++;
		}
		else
		{
			tmp_coord_delete->set1Value(i,diff);
		}
	}


	int delete_it = 0;
	diff = 0;
	//recalculate and delete triangles
	for (i = 0; i < tmp_ifs->coordIndex.getNum() / 4; i++)
	{
		for (j = 0; j < 3; j++)
		{
			tri[j] = tmp_ifs->coordIndex[i * 4 + j];
		}

		delete_it = 0;
		for (j = 0; j < 3; j++)
		{

			if (((*tmp_coord_join)[tri[j]] != -1) && (((*tmp_coord_join)[tri[j]] == tri[(j + 1) % 3]) || ((*tmp_coord_join)[tri[j]] == tri[(j + 2) % 3])))
			{ //triangle to delete
				delete_it = 1;
			}
		}

		//change coordinate
		if (delete_it == 0)
		{
			for (j = 0; j < 3; j++)
			{
				if ((*tmp_coord_join)[tri[j]] != -1)
				{
					tmp_ifs->coordIndex.set1Value(i * 4 + j,(*tmp_coord_join)[tri[j]] - (*tmp_coord_delete)[(*tmp_coord_join)[tri[j]]]);
				}
				else
				{
					//decrement index to coordinate3
					tmp_ifs->coordIndex.set1Value(i * 4 + j,tri[j] - (*tmp_coord_delete)[tri[j]]);
				}
			}
		}
		else
		{
			//delete triangle
			tmp_ifs->coordIndex.deleteValues(i * 4, 4);
			i--;
			diff++;
		}
	}

	//delete coordinates
	diff = 0;
	for (i = 0; i < tmp_coord3->point.getNum(); i++)
	{
		if ((*tmp_coord_join)[i+diff] != -1)
		{ //coordinate to delete
			tmp_coord3->point.deleteValues(i,1);
			tmp_tex->point.deleteValues(i,1);
			diff++;
			i--;
		}
	}
	CorrectTriangles(tmp_coord3,tmp_ifs);

	delete tmp_distance;
	delete tmp_delete;
	delete tmp_join;
	delete tmp_coord_delete;
	delete tmp_coord_join;
	delete tmp_coord_joined;

	return;
}

//! create one LOD segment
SoSeparator * chunkedLOD::CreateLevel(float x1, float y1, float x2, float y2, int level, int maxLevel, int reduce)
{
	int i;
	SoSeparator * dest = new SoSeparator;
	SoSeparator * children = new SoSeparator;
	SoSeparator * me;
	SbBSPTree * border = new SbBSPTree(16,0);

	SoLOD * lod = new SoLOD;

	SoSeparator * child0;
	SoSeparator * child1;
	SoSeparator * child2;
	SoSeparator * child3;

#ifdef MY_DEBUG
	SoSeparator * oddel;
	SoCoordinate3 * c3;
	SoDrawStyle * ds;
	SoPointSet * ps;
#endif

	if (level == maxLevel)
	{
		me = CutSegment(originalModel, x1, y1, x2, y2, border);

#ifdef MY_DEBUG
		oddel = new SoSeparator();

		c3 = new SoCoordinate3();
		c3->point.setNum(border->numPoints());

		ds = new SoDrawStyle();
		ds->pointSize = 10;

		ps = new SoPointSet();
		for (i = 0; i < border->numPoints(); i++)
		{
			c3->point.set1Value(i,border->getPoint(i));
		}

		oddel->addChild(c3);
		oddel->addChild(ds);
		oddel->addChild(ps);
		me->addChild(oddel);
#endif

		//clean memory
		delete border;

		segmentsDone++;
		printf("\rBuilding ChunkedLOD tree ... %d%%", segmentsDone * 100 / totalSegments);
		return me;
	}
	else
	{
		me = CutSegment(originalModel,x1,y1,x2,y2, border);
		child0 = CreateLevel(x1, y1, x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2,level+1,maxLevel,reduce);
		child1 = CreateLevel(x1, y1 + (y2 - y1) / 2, x1 + (x2 - x1) / 2, y2,level+1,maxLevel,reduce);
		child2 = CreateLevel(x1 + (x2 - x1) / 2, y1, x2, y1 + (y2 - y1) / 2,level+1,maxLevel,reduce);
		child3 = CreateLevel(x1 + (x2 - x1) / 2, y1 + (y2 - y1) / 2, x2, y2,level+1,maxLevel,reduce);

		lod->range.set1Value(0, (x2 - x1)*LOD_DISTANCE);
		lod->center.setValue((x1 + x2) / 2,0,(y1 + y2) / 2);
		children->addChild(child0);
		children->addChild(child1);
		children->addChild(child2);
		children->addChild(child3);

		lod->addChild(children);
		for (i = 0; i < maxLevel - level; i++)
		{
			SimplifySegment(me,border,reduce);
			segmentsDone++;
			printf("\rBuilding ChunkedLOD tree ... %d%%", segmentsDone * 100 / totalSegments);
		}

#ifdef MY_DEBUG
		oddel = new SoSeparator();

		c3 = new SoCoordinate3();
		c3->point.setNum(border->numPoints());

		ds = new SoDrawStyle();
		ds->pointSize = 10;

		ps = new SoPointSet();
		for (i = 0; i < border->numPoints(); i++)
		{
			c3->point.set1Value(i,border->getPoint(i));
		}

		oddel->addChild(c3);
		oddel->addChild(ds);
		oddel->addChild(ps);
		me->addChild(oddel);
#endif
		//add to final node
		lod->addChild(me);
		if (level == 1)
		{
/*
			SoTexture2 * tex2 = new SoTexture2;
//			tex2->filename.setValue(".\\strasse.jpg");
			tex2->filename.setValue(".\\cilaos_5%.jpg");
			tex2->blendColor.setValue(0,0,0);
			tex2->model = SoTexture2::REPLACE;
			dest->addChild(tex2);
*/
			AddTexture(dest);
/*
			//shapehints
			SoShapeHints * shape_hints = new SoShapeHints;
			shape_hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
			shape_hints->shapeType = SoShapeHints::UNKNOWN_SHAPE_TYPE;
			shape_hints->faceType = SoShapeHints::UNKNOWN_FACE_TYPE;
			shape_hints->creaseAngle = 0;
			dest->addChild(shape_hints);
*/
		}

		dest->addChild(lod);

		//clean memory
		delete border;

		return dest;
	}
}

//! build chunkedLOD quadtree
SoSeparator * chunkedLOD::BuildChunkedLOD(int maxLevel, int reduce)
{
	float minx, miny, maxx, maxy;
	int tmp;

	segmentsDone = 0;
	totalSegments = maxLevel - 1;
	tmp = 1;
	for (int i = 1; i < maxLevel; i++)
	{
		if (i != maxLevel - 1)
		{
			totalSegments += tmp * 4 * (maxLevel - i - 1);
		}
		else
		{
			totalSegments += tmp * 4;
		}
		tmp = tmp * 4;
	}

	FindMinMax(originalModel,minx, miny, maxx, maxy);

	root = CreateLevel(minx, miny, maxx, maxy,1,maxLevel,reduce);
	root->ref();
	printf("\rBuilding ChunkedLOD tree ... Done\n");
	return root;
}
