/*****************************************************************************\
 *
 * Main-Windows.cpp
 *
 * Main-Windows - main function for Windows platform (SoWin)
 *
 * Authors: Petr "Lamik" Vrba (xvrbap00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 *
\*****************************************************************************/

#include <cstdlib>
#include <string.h> 
#include <Inventor/SoDB.h>
#include <Inventor/sensors/SoOneShotSensor.h>
#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/viewers/SoWinExaminerViewer.h>

#include "chunkedLOD.h"
#include "SoPerfGraph.h"

// Linker: Link with main (not WinMain)
// and use a console
#pragma comment(linker, "/entry:\"mainCRTStartup\"")
#pragma comment(linker, "/subsystem:\"console\"")

//#define MY_DEBUG

bool headlight = true;           // Headlight

char title[50] = "ChunkedLOD algorithm for universal model";

#ifdef MY_DEBUG
SoPerfGraph *perfGraphFPS;

static SbTime lastTime = SbTime::zero();

//! callback function for FPS counting
static void timeTick(void * data, SoSensor * sensor)
{
  SbTime dt;

  if (lastTime == SbTime::zero()) {
    lastTime = SbTime::getTimeOfDay();
    dt = SbTime::zero();
  } else {
    SbTime ct = SbTime::getTimeOfDay();
    dt = ct - lastTime;
    lastTime = ct;
  }

  // update FPS graph
  if (dt.getValue() != 0.)
    perfGraphFPS->appendValue(1.f/float(dt.getValue()), "FPS");

  sensor->schedule();
}
#endif

//! clear console screen
void cls()
{
	HANDLE hOutput;
	DWORD dwWritten;
	COORD coordSize;
	CONSOLE_SCREEN_BUFFER_INFO conInfo;
	DWORD size;

	coordSize.X = 0;
	coordSize.Y = 0;
	
	hOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hOutput, &conInfo);
	size = conInfo.dwSize.X * conInfo.dwSize.Y;
	FillConsoleOutputCharacter(hOutput,' ',size,coordSize,&dwWritten);
	FillConsoleOutputAttribute(hOutput,0,size,coordSize,&dwWritten);
	SetConsoleCursorPosition(hOutput, coordSize);
//	SetConsoleTextAttribute(hOutput,FOREGROUND_GREEN);
}

void welcome()
{
	printf("/------------------------------------------------------------------------------\\");
	printf("|                       Chunked LOD terrain rendering                          |");
	printf("\\------------------------------------------------------------------------------/");
}

//conversion from string to float
#ifdef _MSC_VER
float strtof(const char *nptr, char **endptr)  { return float(strtod(nptr, endptr)); }
#endif

//conversion from string to number
void strtoX(int &r, const char *nptr, char **endptr)
{
  r = strtol(nptr, endptr, 10);
}
void strtoX(float &r, const char *nptr, char **endptr)
{
  r = strtof(nptr, endptr);
}
void strtoX(bool &r, const char *nptr, char **endptr)
{
  int i;
  strtoX(i, nptr, endptr);
  r = (i != 0);
}

//read parameters from commandline
template<class Type>
bool getParam(int argc, char **argv, int &position,
              const char *param, Type &result)
{
  if (argc > position && strcmp(argv[position], param) == 0) {
    position++;
    if (argc > position) {
      char *errp;
      Type tmp;
      strtoX(tmp, argv[position], &errp);
      if (*(argv[position]) != '\0' && *errp == '\0') {
        result = tmp;
        position++;
        return true;
      } else {
        printf("Missing argument \"%s\".", param);
        exit(EXIT_FAILURE);
      }
    } else {
      printf("Missing value for parameter \"%s\".", param);
      exit(EXIT_FAILURE);
    }
  }
  return false;
}

int main(int argc, char **argv)
{
	SoSeparator  * root = NULL;
	SoSeparator  * root2= NULL;
	int i = 2;
	int level = DEFAULT_LEVEL;
	int reduce = DEFAULT_REDUCE;
	int operation = 0;
	float x1 = -1000.0;
	float y1 = -1000.0;
	float x2 = 1000.0;
	float y2 = 1000.0;
	float minx, miny, maxx, maxy;
	SbBSPTree * borders = NULL;
	int dot_pos = 0;
	char filename[1024];

	

	// show FPS
	putenv("COIN_SHOW_FPS_COUNTER=1");

	// switch off sound (console freezing problem)
	putenv("COIN_SOUND_ENABLE=0");
	
	SoDB::init();

  // Initialisation
  HWND window = SoWin::init(argv[0]);
  if (window == NULL) exit(1);


	//clearscreen
	cls();

	//welcome screen
	welcome();

	if (argc < 2)
	{
		printf("Usage: %s filename [-operation int] [-level int] [-reduce int]\n"
			"-operation int    0 = show model, 1 = reduce model and save it to file\n"
			"-level int        max LOD, default %d\n"
			"-reduce int       percentage of triangles in model to eliminate, default %d"
			,argv[0],DEFAULT_LEVEL,DEFAULT_REDUCE);
		exit(EXIT_FAILURE);
	}
  getParam<int>(argc, argv, i, "-operation", operation);
  getParam<int>(argc, argv, i, "-level", level);
  getParam<int>(argc, argv, i, "-reduce", reduce);
  getParam<float>(argc, argv, i, "-x1", x1);
  getParam<float>(argc, argv, i, "-y1", y1);
  getParam<float>(argc, argv, i, "-x2", x2);
  getParam<float>(argc, argv, i, "-y2", y2);

	chunkedLOD chLOD;
	if (level <= 1) level = 2;
	if (level > 10) level = 10;
	if (reduce >= 90) reduce = 90;
	if (reduce < 0) reduce = 0;
	if (operation < 0) operation = 0;
	if (operation > 4) operation = 0;

	sprintf(filename,argv[1]);
	sprintf(strstr(filename,".iv"),".lod.iv");
		
	if (!chLOD.LoadModel(argv[1]))
	{
		exit(EXIT_FAILURE);
	}

	switch (operation)
	{
		case 0:
			root = chLOD.BuildChunkedLOD(level,reduce);
			break;

		case 1:
			root = chLOD.getOriginalModel();
			break;

		case 2:
			borders = new SbBSPTree(64, 0);
			root = new SoSeparator();
			root->ref();


			chLOD.AddTexture(root);

			printf("Reducing model ...");
			chLOD.FindMinMax(chLOD.getOriginalModel(),minx, miny, maxx, maxy);
			root2= chLOD.CutSegment(chLOD.getOriginalModel(),minx, miny, maxx, maxy, borders);
			chLOD.SimplifySegment(root2, borders, reduce);
			printf("Done\n");

			root->addChild(root2);

			//create output filename
			sprintf(filename,argv[1]);
			sprintf(strstr(filename,".iv"),".reduced.iv");

			//save model to file
			chLOD.SaveModel(root,filename);


			delete borders;
			break;
		case 3:
			borders = new SbBSPTree(64, 0);
	
			root = new SoSeparator();
			root->ref();
			chLOD.AddTexture(root);
			printf("Cutting model ...");
			root->addChild(chLOD.CutSegment(chLOD.getOriginalModel(), x1, y1, x2, y2, borders));
			printf("Done\n");

			delete borders;

			//create output filename
			sprintf(filename,argv[1]);
			sprintf(strstr(filename,".iv"),".noRECT.iv");

			//save model to file
			chLOD.SaveModel(root,filename);

			break;
	}

#ifdef MY_DEBUG
  SoOneShotSensor *timeSensor = new SoOneShotSensor(timeTick, NULL);
  timeSensor->schedule();
  
	//performance monitor
  SoPerfGraph::initClass();
  perfGraphFPS = new SoPerfGraph;
  perfGraphFPS->horScreenOrigin = SoPerfGraph::RIGHT;
  perfGraphFPS->horAlignment = SoPerfGraph::RIGHT;
  perfGraphFPS->position.setValue(-0.03f,0.03f);
  root->addChild(perfGraphFPS); 
#endif

  // Create window
  SoWinExaminerViewer *viewer = new SoWinExaminerViewer(window);
  viewer->setSceneGraph(root);
  viewer->setHeadlight(headlight);
  viewer->setTitle(title);

  viewer->show();
 
  // Show window and start render loop
  SoWin::show(window);
  SoWin::mainLoop();

  // Free memory
#ifdef MY_DEBUG
  delete timeSensor;
#endif
  delete viewer;
  root->unref();
#ifdef MY_DEBUG
  SoPerfGraph::cleanup();
#endif

  return 0;

}

